"""
XML representers.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 19, 2011.
"""
from collections import OrderedDict
import datetime

from lxml import etree
from lxml import objectify
from pkg_resources import resource_filename # pylint: disable=E0611
from pyramid.compat import bytes_
from pyramid.compat import text_type

from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.constants import RESOURCE_KINDS
from everest.mime import XmlMime
from everest.representers.base import MappingResourceRepresenter
from everest.representers.base import RepresentationGenerator
from everest.representers.base import RepresentationParser
from everest.representers.config import RepresenterConfiguration
from everest.representers.converters import BooleanConverter
from everest.representers.converters import ConverterRegistry
from everest.representers.converters import DateTimeConverter
from everest.representers.dataelements import CollectionDataElement
from everest.representers.dataelements import LinkedDataElement
from everest.representers.dataelements import MemberDataElement
from everest.representers.interfaces import ILinkedDataElement
from everest.representers.mapping import MappingRegistry
from everest.representers.utils import get_mapping_registry
from everest.resources.link import Link
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_member_class
from everest.resources.utils import provides_member_resource
from everest.resources.utils import resource_to_url
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['XmlCollectionDataElement',
           'XmlConverterRegistry',
           'XmlLinkedDataElement',
           'XmlMappingRegistry',
           'XmlMemberDataElement',
           'XmlParserFactory',
           'XmlRepresentationGenerator',
           'XmlRepresentationParser',
           'XmlRepresenterConfiguration',
           'XmlResourceRepresenter',
           ]


XML_NS_OPEN_SEARCH = 'http://a9.com/-/spec/opensearch/1.1/'
XML_NS_XSI = 'http://www.w3.org/2001/XMLSchema-instance'

XML_TAG_OPTION = 'xml_tag'
XML_SCHEMA_OPTION = 'xml_schema'
XML_NAMESPACE_OPTION = 'xml_ns'
XML_PREFIX_OPTION = 'xml_prefix'
XML_VALIDATE_OPTION = 'xml_validate'

NAMESPACE_MAPPING_OPTION = 'namespace'


class XmlConverterRegistry(ConverterRegistry):
    pass

XmlConverterRegistry.register(datetime.datetime, DateTimeConverter)
XmlConverterRegistry.register(bool, BooleanConverter)


class XmlRepresentationParser(RepresentationParser):
    def run(self):
        # Create an XML schema.
        schema_loc = self.get_option('schema_location')
        parser = XmlParserFactory.create(schema_location=schema_loc)
        try:
            tree = objectify.parse(self._stream, parser)
        except etree.XMLSyntaxError as err:
            msg = 'Could not parse XML document'
            if not schema_loc is None:
                msg += ' for schema %s.' % schema_loc
            raise SyntaxError('%s\n%s' % (msg, err.msg))
        return tree.getroot()[0]


class XmlRepresentationGenerator(RepresentationGenerator):
    def run(self, data_element):
        objectify.deannotate(data_element)
        etree.cleanup_namespaces(data_element)
        rpr_text = etree.tostring(data_element,
                                  pretty_print=True,
                                  encoding=text_type)
        self._stream.write(rpr_text)


class XmlParserFactory(object):
    __parser = None

    @classmethod
    def create(cls, schema_location=None):
        if not schema_location is None:
            schema = cls.__get_xml_schema(schema_location)
            parser = objectify.makeparser(schema=schema)
        else:
            parser = objectify.makeparser()
        # Get the class lookup from the mapping registry.
        mp_reg = get_mapping_registry(XmlMime)
        parser.set_element_class_lookup(mp_reg.parsing_lookup)
        return parser

    @classmethod
    def __get_xml_schema(cls, xml_schema_path):
        try:
            doc = etree.parse(resource_filename(*xml_schema_path.split(':')))
        except etree.XMLSyntaxError as err:
            raise SyntaxError('Could not parse XML schema %s.\n%s' %
                              (xml_schema_path, err.msg))
        try:
            schema = etree.XMLSchema(doc)
        except etree.XMLSchemaParseError as err:
            raise SyntaxError('Invalid XML schema.\n Parser message: %s'
                              % str(err))
        return schema


class XmlResourceRepresenter(MappingResourceRepresenter):
    content_type = XmlMime

    __tmpl = "<?xml version='1.0' encoding='%s' ?>%s"

    def to_bytes(self, obj, encoding=None):
        """
        Overwritten so we can insert the `?xml` processing directive.
        """
        if encoding is None:
            encoding = self.encoding
        text = self.__tmpl % (encoding, self.to_string(obj))
        return bytes_(text, encoding=encoding)

    def data_to_bytes(self, data_element, encoding=None):
        """
        Overwritten so we can insert the `?xml` processing directive.
        """
        if encoding is None:
            encoding = self.encoding
        text = self.__tmpl % (encoding, self.data_to_string(data_element))
        return bytes_(text, encoding=encoding)

    @classmethod
    def make_mapping_registry(cls):
        return XmlMappingRegistry()

    def _make_representation_parser(self, stream, resource_class, mapping):
        parser = XmlRepresentationParser(stream, resource_class, mapping)
        mp = self._mapping
        do_validate = mp.configuration.get_option(XML_VALIDATE_OPTION)
        if do_validate:
            xml_schema = mp.configuration.get_option(XML_SCHEMA_OPTION)
        else:
            xml_schema = None
        parser.set_option('schema_location', xml_schema)
        return parser

    def _make_representation_generator(self, stream, resource_class, mapping):
        return XmlRepresentationGenerator(stream, resource_class, mapping)


class _XmlDataElementMixin(object):
    @classmethod
    def create(cls, ns_map=None):
        return cls._create(ns_map)

    @classmethod
    def create_from_resource(cls, resource, ns_map=None): # ignore resource pylint:disable=W0613,W0221
        return cls._create(ns_map)

    @classmethod
    def _create(cls, ns_map):
        if ns_map is None:
            mp_reg = get_mapping_registry(XmlMime)
            ns_map = mp_reg.namespace_map
        cls_xml_tag = cls.mapping.configuration.get_option(XML_TAG_OPTION)
        if cls_xml_tag is None:
            raise ValueError('No XML tag registered for mapped class '
                             '%s.' % cls.mapping.mapped_class)
        cls_xml_ns = \
                cls.mapping.configuration.get_option(XML_NAMESPACE_OPTION)
        if not cls_xml_ns is None:
            tag = "{%s}%s" % (cls_xml_ns, cls_xml_tag)
            # FIXME: is this really necessary?
            ns_map[None] = cls_xml_ns
        else:
            tag = cls_xml_tag
        el_fac = XmlParserFactory.create().makeelement
        return el_fac(tag, nsmap=ns_map)


class XmlMemberDataElement(objectify.ObjectifiedElement,
                           _XmlDataElementMixin, MemberDataElement):
    #: This is used to identify nested attributes with None value.
    __NONE_MARKER = '__none__'
    converter_registry = XmlConverterRegistry

    def get_nested(self, attr):
        return self.__get_attribute(attr, True)

    def set_nested(self, attr, data_element):
        self.__set_attribute(attr, data_element)

    def get_terminal(self, attr):
        return self.__get_attribute(attr, True)

    def set_terminal(self, attr, value):
        self.__set_attribute(attr, value)

    def iterator(self):
        id_val = self.get('id')
        yield ('id', None) if id_val is None \
            else ('id', self.mapping.get_attribute('id').value_type(id_val))
        for child in self.iterchildren():
            idx = child.tag.find('}')
            if idx != -1:
                tag = child.tag[idx + 1:]
            else:
                tag = child.tag
            attr = self.mapping.get_attribute_by_repr(tag)
            if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                value = self.__check_for_link(child)
            else:
                value = XmlConverterRegistry.convert_from_representation(
                                                            child.text,
                                                            attr.value_type)
            yield (tag, value)

    @property
    def data(self):
        data_map = OrderedDict()
        for (name, value) in self.iterator():
            data_map[name] = value
        return data_map

    def get_attribute(self, attr_name):
        attr = self.mapping.get_attribute_by_repr(attr_name)
        return self.__get_attribute(attr, False)

    def set_attribute(self, attr_name, value):
        attr = self.mapping.get_attribute_by_repr(attr_name)
        self.__set_attribute(attr, value)

    def __get_q_tag(self, attr):
        # FIXME: We should cache the namespace for each attribute.
        if not attr.namespace is None:
            q_tag = '{%s}%s' % (attr.namespace, attr.repr_name)
        else:
            if attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                xml_ns = \
                  self.mapping.configuration.get_option(XML_NAMESPACE_OPTION)
            else:
                if attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
                    attr_type = get_member_class(attr.value_type)
                elif attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
                    attr_type = get_collection_class(attr.value_type)
                mp = self.mapping.mapping_registry.find_mapping(attr_type)
                if not mp is None:
                    xml_ns = mp.configuration.get_option(XML_NAMESPACE_OPTION)
                else: # pragma: no cover
                    # Not mapped.
                    # FIXME This case is neither tested nor documented.
                    xml_ns = None
            if not xml_ns is None:
                q_tag = '{%s}%s' % (xml_ns, attr.repr_name)
            else:
                q_tag = '{}%s' % attr.repr_name
        return q_tag

    def __get_attribute(self, attr, safe):
        if attr.repr_name == 'id':
            # The "special" id attribute.
            xml_val = self.get('id')
            if not xml_val is None:
                value = attr.value_type(xml_val)
            else:
                value = None
        else:
            q_tag = self.__get_q_tag(attr)
            child_it = self.iterchildren(q_tag)
            try:
                val_el = next(child_it)
            except StopIteration:
                if safe:
                    value = None
                else:
                    raise AttributeError(attr.repr_name)
            else:
                try:
                    next(child_it)
                except StopIteration:
                    pass
                else:
                    # This should never happen.
                    raise ValueError('More than one child for member '
                                     'attribute "%s" found.' % attr) # pragma: no cover
                if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                    if val_el.text == self.__NONE_MARKER:
                        value = None
                    else:
                        value = self.__check_for_link(val_el)
                else:
                    value = XmlConverterRegistry.convert_from_representation(
                                                            val_el.text,
                                                            attr.value_type)
#            q_tag = self.__get_q_tag(attr)
#            val_el = getattr(self, q_tag, None)
#            if not val_el is None:
#                value = XmlConverterRegistry.convert_from_representation(
#                                                            val_el.text,
#                                                            attr.value_type)
#            else:
#                value = None
        return value

    def __set_attribute(self, attr, value):
        if attr.repr_name == 'id':
            # The "special" id attribute.
            self.set('id', str(value))
        else:
            q_tag = self.__get_q_tag(attr)
            if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                if value is None:
                    setattr(self, q_tag, self.__NONE_MARKER)
                else:
                    if not isinstance(value, _XmlDataElementMixin):
                        raise ValueError('Non-terminal attribute value must '
                                         'be None or XML data element.')
                    value.tag = q_tag
                    self.append(value)
            else:
                xml_value = XmlConverterRegistry.convert_to_representation(
                                                            value,
                                                            attr.value_type)
                setattr(self, q_tag, xml_value)

    def __check_for_link(self, child):
        # Link handling: look for wrapper tag with *one* link child.
        if child.countchildren() == 1:
            grand_child = child.getchildren()[0]
            if ILinkedDataElement in provided_by(grand_child):
#                # We inject the id attribute from the wrapper element.
#                str_xml = child.get('id')
#                if not str_xml is None:
#                    grand_child.set('id', str_xml)
                child = grand_child
        return child


class XmlCollectionDataElement(objectify.ObjectifiedElement,
                               _XmlDataElementMixin, CollectionDataElement):
    def add_member(self, data_element):
        self.append(data_element)

    def get_members(self):
        return self.iterchildren()

    def __len__(self):
        return self.countchildren()


class XmlLinkedDataElement(objectify.ObjectifiedElement, LinkedDataElement):
    @classmethod
    def create(cls, url, kind,
               id=None, relation=None, title=None, **options): # pylint: disable=W0622
#        mp_reg = get_mapping_registry(XmlMime)
#        ns_map = mp_reg.namespace_map
        xml_ns = options[XML_NAMESPACE_OPTION]
        el_fac = XmlParserFactory.create().makeelement
        tag = '{%s}link' % xml_ns
        link_el = el_fac(tag)
        link_el.set('href', url)
        link_el.set('kind', kind)
        if not id is None:
            link_el.set('id', id)
        if not relation is None:
            link_el.set('rel', relation)
        if not title is None:
            link_el.set('title', title)
        return link_el

    @classmethod
    def create_from_resource(cls, resource):
        # Create the wrapping element.
        mp_reg = get_mapping_registry(XmlMime)
        mp = mp_reg.find_or_create_mapping(type(resource))
        xml_ns = mp.configuration.get_option(XML_NAMESPACE_OPTION)
        options = {XML_NAMESPACE_OPTION:xml_ns}
        rc_data_el = mp.create_data_element_from_resource(resource)
        if provides_member_resource(resource):
            link_el = cls.create(resource_to_url(resource),
                                 RESOURCE_KINDS.MEMBER,
                                 id=str(resource.id),
                                 relation=resource.relation,
                                 title=resource.title,
                                 **options)
            rc_data_el.set('id', str(resource.id))
            rc_data_el.append(link_el)
        else: # collection resource.
            # Collection links only get an actual link element if they
            # contain any members.
            link_el = cls.create(resource_to_url(resource),
                                 RESOURCE_KINDS.COLLECTION,
                                 relation=resource.relation,
                                 title=resource.title,
                                 **options)
            rc_data_el.append(link_el)
        return rc_data_el

    def get_url(self):
        return self.get('href')

    def get_kind(self):
        return self.get('kind')

    def get_relation(self):
        return self.get('rel')

    def get_title(self):
        return self.get('title')

    def get_id(self):
        # FIXME: This will not work with ID strings that happen to be
        #        convertible to an int.
        id_str = self.get('id')
        try:
            id_val = int(id_str)
        except ValueError:
            id_val = id_str
        except TypeError: # Happens if the id string is None.
            id_val = id_str
        return id_val


class XmlRepresenterConfiguration(RepresenterConfiguration):
    """
    Specialized configuration class for XML representers.

    Allowed configuration attribute names:

    xml_tag :
        The XML tag to use for the represented data element class.
    xml_schema :
        The XML schema to use for the represented data element class.
    xml_ns :
        The XML namespace to use for the represented data element class.
    xml_prefix :
        The XML namespace prefix to use for the represented data element
        class.
    xml_validate:
        Boolean flag indicating if incoming representations should be
        validated upon parsing (defaults to `True`).
    """
    _default_config_options = \
        dict(list(RepresenterConfiguration._default_config_options.items())
             + [(XML_TAG_OPTION, None), (XML_SCHEMA_OPTION, None),
                (XML_NAMESPACE_OPTION, None), (XML_PREFIX_OPTION, None),
                (XML_VALIDATE_OPTION, True)])
    _default_attributes_options = \
        dict(list(RepresenterConfiguration._default_attributes_options.items())
             + [(NAMESPACE_MAPPING_OPTION, None)])


class XmlMappingRegistry(MappingRegistry):
    """
    Registry for XML mappings.
    """

    member_data_element_base_class = XmlMemberDataElement
    collection_data_element_base_class = XmlCollectionDataElement
    linked_data_element_base_class = XmlLinkedDataElement
    configuration_class = XmlRepresenterConfiguration

    #: Static namespace prefix: namespace map.
    NS_MAP = dict(xsi=XML_NS_XSI)

    def __init__(self):
        MappingRegistry.__init__(self)
        self.__ns_map = self.__class__.NS_MAP.copy()
        self.__ns_lookup = None

    def _initialize(self):
        # Create and register the linked data element class.
        configuration = self.configuration_class()
        mapping = self.create_mapping(Link, configuration)
        self.set_mapping(mapping)

    def set_mapping(self, mapping):
        # First, record the namespace and prefix, if necessary.
        xml_ns = mapping.configuration.get_option(XML_NAMESPACE_OPTION)
        if not xml_ns is None:
            xml_prefix = mapping.configuration.get_option(XML_PREFIX_OPTION)
            if not xml_prefix is None:
                ns = self.__ns_map.get(xml_prefix)
                if ns is None:
                    # New prefix - register.
                    self.__ns_map[xml_prefix] = xml_ns
                elif xml_ns != ns:
                    raise ValueError('Prefix "%s" is already registered for '
                                     'namespace %s.' % (xml_prefix, ns))
        # Make sure we rebuild the lookup.
        if not self.__ns_lookup is None:
            self.__ns_lookup = None
        MappingRegistry.set_mapping(self, mapping)

    @property
    def namespace_map(self):
        return self.__ns_map.copy()

    @property
    def parsing_lookup(self):
        if self.__ns_lookup is None:
            self.__ns_lookup = self.__create_parsing_lookup()
        return self.__ns_lookup

    def __create_parsing_lookup(self):
        lookup = etree.ElementNamespaceClassLookup(
                                objectify.ObjectifyElementClassLookup())
        for mapping in self.get_mappings():
            de_cls = mapping.data_element_class
            if issubclass(de_cls, XmlLinkedDataElement):
                continue
            xml_ns = mapping.configuration.get_option(XML_NAMESPACE_OPTION)
            xml_tag = bytes_(mapping.configuration.get_option(XML_TAG_OPTION))
            ns_cls_map = lookup.get_namespace(xml_ns)
            if xml_tag in ns_cls_map:
                raise ValueError('Duplicate tag "%s" in namespace "%s" '
                                 '(trying to register class %s)'
                                 % (xml_tag, xml_ns, de_cls))
            ns_cls_map[xml_tag] = de_cls
            ns_cls_map['link'] = XmlLinkedDataElement
        return lookup



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 9, 2012.
"""
import pytest

from everest.mime import CsvMime
from everest.mime import XmlMime
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import WRITE_AS_LINK_OPTION
from everest.representers.dataelements import DataElementAttributeProxy
from everest.representers.dataelements import LinkedDataElement
from everest.representers.interfaces import IRepresenterRegistry
from everest.representers.utils import get_mapping_registry
from everest.representers.xml import XML_NAMESPACE_OPTION
from everest.representers.xml import XML_TAG_OPTION
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.resources import MyEntityMember
from everest.tests.complete_app.testing import create_collection
from everest.tests.complete_app.testing import create_entity
from everest.tests.test_entities import MyEntity
from everest.mime import JsonMime


__docformat__ = 'reStructuredText en'
__all__ = ['TestMapping',
           ]


class TestMapping(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_rpr.zcml'

    def test_defaults(self, mapping_registry_factory):
        mp = mapping_registry_factory(CsvMime).find_or_create_mapping(
                                                                MyEntityMember)
        attrs = mp.get_attribute_map()
        assert attrs['text'].options.get(IGNORE_OPTION) is None
        assert attrs['parent'].options.get(IGNORE_OPTION) is False
        key = ('parent',)
        parent_attrs = mp.get_attribute_map(key=key)
        assert parent_attrs['text'].options.get(IGNORE_OPTION) is None

    def test_clone_update_attributes(self, mapping_registry_factory):
        key = ('parent', 'text')
        mp = mapping_registry_factory(CsvMime).find_or_create_mapping(
                                                                MyEntityMember)
        mp1 = mp.clone(attribute_options={key:{IGNORE_OPTION:True}})
        p_key = ('parent',)
        parent_attrs = mp1.get_attribute_map(key=p_key)
        assert parent_attrs['text'].options.get(IGNORE_OPTION) is True
        mp1.update(attribute_options={key:{IGNORE_OPTION:False}})
        parent_attrs_upd = mp1.get_attribute_map(key=p_key)
        assert parent_attrs_upd['text'].options.get(IGNORE_OPTION) is False

    @pytest.mark.parametrize('mime,rpr_attr',
                             [(CsvMime, 'parent'),
                              (XmlMime, 'myentityparent'),
                              (JsonMime, 'parent')
                              ])
    def test_attribute_access(self, mapping_registry_factory, mime, rpr_attr):
        #
        mp = mapping_registry_factory(mime).find_or_create_mapping(
                                                                MyEntityMember)
        assert mp.has_attribute('parent')
        assert mp.has_attribute_repr(rpr_attr)
        attr = mp.get_attribute('parent')
        assert attr is mp.get_attribute_map()['parent']
        with pytest.raises(AttributeError):
            dummy = mp.get_attribute('prnt')

    def test_map_to_data_element(self, resource_repo):
        def _test(mb, cnt_type, parent_repr_name, children_repr_name):
            mp_reg = get_mapping_registry(cnt_type)
            mp = mp_reg.find_or_create_mapping(MyEntityMember)
            de = mp.map_to_data_element(mb)
            prx = DataElementAttributeProxy(de)
            assert prx.get_data_element() is de
            assert prx.id == 0
            assert prx.text == 'TEXT'
            assert prx.number == 1
            if cnt_type is XmlMime:
                # The parent attribute is a link.
                assert isinstance(getattr(prx, parent_repr_name),
                                  LinkedDataElement)
                with pytest.raises(AttributeError):
                    dummy = getattr(prx, children_repr_name)
            else:
                assert isinstance(getattr(prx, parent_repr_name),
                                  DataElementAttributeProxy)
                children_el = getattr(prx, children_repr_name)
                assert isinstance(children_el, LinkedDataElement)
            # Nonexisting attribute raises error.
            with pytest.raises(AttributeError):
                dummy = getattr(prx, 'foo')
            with pytest.raises(AttributeError):
                setattr(prx, 'foo', 'murks')
            # Set terminal attribute.
            prx.id = 1
            assert prx.id == 1
            # Set nested attribute.
            setattr(prx, parent_repr_name, None)
            assert getattr(prx, parent_repr_name) is None
            with pytest.raises(ValueError):
                setattr(prx, parent_repr_name, 1)
        entity = create_entity()
        coll = resource_repo.get_collection(IMyEntity)
        mb = coll.create_member(entity)
        _test(mb, XmlMime, 'myentityparent', 'myentitychildren')
        _test(mb, CsvMime, 'parent', 'children')

    def test_map_to_data_element_with_member(self, resource_repo):
        entity = create_entity()
        coll = resource_repo.get_collection(IMyEntity)
        mb = coll.create_member(entity)
        mp_reg = get_mapping_registry(CsvMime)
        mp = mp_reg.find_or_create_mapping(MyEntityMember)
        mp1 = mp.clone(
            attribute_options={('parent',):{WRITE_AS_LINK_OPTION:False},
                             })
        de = mp1.map_to_data_element(mb)
        prx = DataElementAttributeProxy(de)
        assert not prx.parent is None
        #
        entity.parent = None
        de1 = mp1.map_to_data_element(mb)
        prx1 = DataElementAttributeProxy(de1)
        with pytest.raises(AttributeError):
            dummy = getattr(prx1, 'parent')

    def test_map_to_data_element_with_collection(self, resource_repo):
        entity = create_entity()
        coll = resource_repo.get_collection(IMyEntity)
        mb = coll.create_member(entity)
        assert len(entity.children) == 1
        assert len(mb.children) == 1
        mb_child = next(iter(mb.children))
        assert len(mb_child.children) == 1
        mp_reg = get_mapping_registry(CsvMime)
        mp = mp_reg.find_or_create_mapping(MyEntityMember)
        mp1 = mp.clone(
            attribute_options={('children',):{IGNORE_OPTION:False,
                                            WRITE_AS_LINK_OPTION:False},
                               ('children', 'children'):
                                             {IGNORE_OPTION:False,
                                              WRITE_AS_LINK_OPTION:False}
                                })
        de = mp1.map_to_data_element(mb)
        prx = DataElementAttributeProxy(de)
        assert len(prx.children) == 1
        assert len(prx.children[0].children) == 1

    def test_mapping_duplicate_prefix(self, new_configurator):
        coll_cls = new_configurator.registry.getUtility(IMyEntity,
                                                    name='collection-class')
        rpr_reg = new_configurator.registry.queryUtility(IRepresenterRegistry)
        mp_reg = rpr_reg.get_mapping_registry(XmlMime)
        mp = mp_reg.find_or_create_mapping(coll_cls)
        ns = 'foo'
        mp.configuration.set_option(XML_NAMESPACE_OPTION, ns)
        with pytest.raises(ValueError) as cm:
            mp.mapping_registry.set_mapping(mp)
        exc_msg = 'is already registered for namespace'
        assert str(cm.value).find(exc_msg) != -1

    def test_mapping_duplicate_tag(self, new_configurator):
        coll_cls = new_configurator.registry.getUtility(IMyEntity,
                                                name='collection-class')
        rpr_reg = new_configurator.registry.queryUtility(IRepresenterRegistry)
        mp_reg = rpr_reg.get_mapping_registry(XmlMime)
        mp = mp_reg.find_or_create_mapping(coll_cls)
        mb_mp = mp_reg.find_or_create_mapping(MyEntityMember)
        mb_tag = mb_mp.configuration.get_option(XML_TAG_OPTION)
        mp.configuration.set_option(XML_TAG_OPTION, mb_tag)
        mp.mapping_registry.set_mapping(mp)
        with pytest.raises(ValueError) as cm:
            getattr(mp.mapping_registry, 'parsing_lookup')
        assert str(cm.value).startswith('Duplicate tag')

    def test_mapping_reset_lookup(self, new_configurator):
        coll_cls = new_configurator.registry.getUtility(IMyEntity,
                                                    name='collection-class')
        rpr_reg = new_configurator.registry.queryUtility(IRepresenterRegistry)
        mp_reg = rpr_reg.get_mapping_registry(XmlMime)
        old_lookup = mp_reg.parsing_lookup
        mp = mp_reg.find_or_create_mapping(coll_cls)
        new_tag = 'my-new-entities'
        mp.configuration.set_option(XML_TAG_OPTION, new_tag)
        mp_reg.set_mapping(mp)
        new_lookup = mp_reg.parsing_lookup
        assert not old_lookup is new_lookup
        ns = mp.configuration.get_option(XML_NAMESPACE_OPTION)
        cls_map = new_lookup.get_namespace(ns)
        assert cls_map[new_tag] == mp.data_element_class

    def test_mapping_linked_xml_data_element_with_string_id(self,
                                                mapping_registry_factory,
                                                resource_repo): # pylint: disable=W0613
        # FIXME: Using resource_repo only for its side effect here.
        mp_reg = mapping_registry_factory(XmlMime)
        mb_mp = mp_reg.find_or_create_mapping(MyEntityMember)
        coll = create_collection()
        mb = next(iter(coll))
        mb_id = 'unique'
        mb.id = mb_id
        data_el = mb_mp.create_linked_data_element_from_resource(mb)
        link_el = next(data_el.iterchildren())
        assert link_el.get_id() == mb.id

    def test_mapping_polymorhpic(self, new_configurator,
                                 mapping_registry_factory):
        coll_cls = new_configurator.registry.getUtility(IMyEntity,
                                                        name='collection-class')
        # pylint: disable=W0232
        class IMyDerivedEntity(IMyEntity):
            pass
        class MyDerivedEntity(MyEntity):
            pass
        class MyDerivedEntityMember(MyEntityMember):
            pass
        class MyDerivedEntityCollection(coll_cls):
            pass
        # pylint: enable=W0232
        new_configurator.add_resource(IMyDerivedEntity, MyDerivedEntityMember,
                                      MyDerivedEntity,
                                      MyDerivedEntityCollection,
                                      expose=False)
        new_configurator.add_resource_representer(
                                            IMyDerivedEntity,
                                            XmlMime,
                                            attribute_options=
                                            {('parent',):dict(ignore=True)})
        new_configurator.begin()
        try:
            mp_reg = mapping_registry_factory(XmlMime)
            mp = mp_reg.find_or_create_mapping(coll_cls)
            for rc in (MyDerivedEntityMember, MyDerivedEntityCollection):
                attr = None
                for attr in mp.attribute_iterator(rc):
                    if attr.name == 'parent':
                        break
                assert not attr is None
                assert getattr(attr, IGNORE_OPTION) is True
        finally:
            new_configurator.end()



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 11, 2013.
"""
import pytest

from everest.constants import CARDINALITIES
from everest.constants import CARDINALITY_CONSTANTS
from everest.constants import Cardinality


__docformat__ = 'reStructuredText en'
__all__ = ['TestConstants',
           ]


class TestConstants(object):
    def test_iterable(self):
        assert set(CARDINALITY_CONSTANTS) == \
                          set((CARDINALITY_CONSTANTS.ONE,
                               CARDINALITY_CONSTANTS.MANY))

    def test_new_cardinality(self):
        with pytest.raises(ValueError) as cm:
            dummy = Cardinality('foo', 'bar')
        assert str(cm.value).startswith('"relator" and')

    def test_str_cardinality(self):
        assert str(CARDINALITIES.MANYTOONE).startswith('*->')



"""
General purpose utilities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 7, 2011.
"""
from collections import MutableSet
from everest.querying.interfaces import IFilterSpecificationVisitor
from everest.querying.interfaces import IOrderSpecificationVisitor
from everest.repositories.interfaces import IRepositoryManager
from functools import update_wrapper
from pyramid.compat import NativeIO
from pyramid.compat import iteritems_
from pyramid.threadlocal import get_current_registry
from weakref import WeakKeyDictionary
from weakref import ref
import re
import traceback
from logging import Formatter
import functools

__docformat__ = 'reStructuredText en'
__all__ = ['BidirectionalLookup',
           'EMAIL_REGEX',
           'WeakList',
           'WeakOrderedSet',
           'check_email',
           'classproperty',
           'generative',
           'get_filter_specification_visitor',
           'get_nested_attribute',
           'get_order_specification_visitor',
           'get_repository_manager',
           'get_traceback',
           'id_generator',
           'resolve_nested_attribute',
           'set_nested_attribute',
           ]


EMAIL_REGEX = r'^[a-zA-Z0-9._%+-]+@(?:[a-zA-Z0-9-]+\.)+[a-zA-Z]{2,4}$'

check_email = re.compile(EMAIL_REGEX).match


class classproperty(object):
    """
    Property descriptor for class objects.
    """

    def __init__(self, get):
        self.__get = get

    def __get__(self, instance, cls):
        return self.__get(cls)


def resolve_nested_attribute(obj, attribute):
    #: Helper function for dotted attribute resolution.
    tokens = attribute.split('.')
    for token in tokens[:-1]:
        obj = getattr(obj, token)
        if obj is None:
            break
    return (obj, tokens[-1])


def get_nested_attribute(obj, attribute):
    """
    Returns the value of the given (possibly dotted) attribute for the given
    object.

    If any of the parents on the nested attribute's name path are `None`, the
    value of the nested attribute is also assumed as `None`.

    :raises AttributeError: If any attribute access along the attribute path
      fails with an `AttributeError`.
    """
    parent, attr = resolve_nested_attribute(obj, attribute)
    if not parent is None:
        attr_value = getattr(parent, attr)
    else:
        attr_value = None
    return attr_value


def set_nested_attribute(obj, attribute, value):
    """
    Sets the value of the given (possibly dotted) attribute for the given
    object to the given value.

    :raises AttributeError: If any of the parents on the nested attribute's
      name path are `None`.
    """
    parent, attr = resolve_nested_attribute(obj, attribute)
    if parent is None:
        raise AttributeError('Can not set attribute "%s" on None value.'
                             % attr)
    setattr(parent, attr, value)


def id_generator(start=0):
    """
    Generator for sequential numeric numbers.
    """
    count = start
    while True:
        send_value = (yield count)
        if not send_value is None:
            if send_value < count:
                raise ValueError('Values from ID generator must increase '
                                 'monotonically (current value: %d; value '
                                 'sent to generator: %d).'
                                 % (count, send_value))
            count = send_value
        else:
            count += 1


def get_traceback():
    """
    Fetches the last traceback from :var:`sys.exc_info` and returns it as a
    formatted string.

    :returns: formatted traceback (string)
    """
    buf = NativeIO()
    traceback.print_exc(file=buf)
    return buf.getvalue()


def get_filter_specification_visitor(name):
    """
    Returns a the class registered as the filter specification
    visitor utility under the given name (one of the
    :const:`everest.querying.base.EXPRESSION_KINDS` constants).

    :returns: class implementing
        :class:`everest.interfaces.IFilterSpecificationVisitor`
    """
    reg = get_current_registry()
    return reg.getUtility(IFilterSpecificationVisitor, name=name)


def get_order_specification_visitor(name):
    """
    Returns the class registered as the order specification
    visitor utility under the given name (one of the
    :const:`everest.querying.base.EXPRESSION_KINDS` constants).

    :returns: class implementing
        :class:`everest.interfaces.IOrderSpecificationVisitor`
    """
    reg = get_current_registry()
    return reg.getUtility(IOrderSpecificationVisitor, name=name)


def get_repository_manager():
    """
    Registers the object registered as the repository manager utility.

    :returns: object implementing
        :class:`everest.interfaces.IRepositoryManager`
    """
    reg = get_current_registry()
    return reg.getUtility(IRepositoryManager)


def get_repository(name=None):
    """
    Returns the repository with the given name or the default repository if
    :param:`name` is `None`.
    """
    repo_mgr = get_repository_manager()
    if name is None:
        repo = repo_mgr.get_default()
    else:
        repo = repo_mgr.get(name)
    return repo


class BidirectionalLookup(object):
    """
    Bidirectional mapping between a left and a right collection of items.

    Each element of the left collection is mapped to exactly one element of
    the right collection; both collections contain unique elements.
    """

    def __init__(self, init_map=None, map_type=dict):
        """
        :param init_map: map-like object to initialize this instance with
        :param map_type: type to use for the left and right item maps
          (dictionary like)
        """
        self.__left = map_type()
        self.__right = map_type()
        if not init_map is None:
            for left, right in iteritems_(init_map):
                self.__setitem__(left, right)

    def __setitem__(self, left_item, right_item):
        is_right_in_left = right_item in self.__left
        is_left_in_right = left_item in self.__right
        in_both = is_right_in_left and is_left_in_right
        if is_right_in_left or is_left_in_right and not in_both:
            raise ValueError('Cannot use the existing item "%s" from the %s '
                             'set as item in the %s set!'
                             % ((left_item, 'right', 'left') if
                                is_left_in_right else
                                (right_item, 'left', 'right')))
        if left_item in self.__left:
            del self.__right[self.__left[left_item]]
        self.__left[left_item] = right_item
        if right_item in self.__right:
            del self.__left[self.__right[right_item]]
        self.__right[right_item] = left_item

    def __delitem__(self, left_or_right_item):
        if left_or_right_item in self.__left:
            del self.__left[left_or_right_item]
        elif left_or_right_item in self.__right:
            del self.__right[left_or_right_item]
        else:
            raise KeyError("'%s'" % (left_or_right_item,))

    def __getitem__(self, left_or_right_item):
        if left_or_right_item in self.__left:
            return self.__left[left_or_right_item]
        elif left_or_right_item in self.__right:
            return self.__right[left_or_right_item]
        else:
            raise KeyError("'%s'" % (left_or_right_item,))

    def __contains__(self, left_or_right_item):
        return (left_or_right_item in self.__left) \
               or (left_or_right_item in self.__right)

    def has_left(self, item):
        return item in self.__left

    def has_right(self, item):
        return item in self.__right

    def get(self, left_or_right_item):
        try:
            item = self.__getitem__(left_or_right_item)
        except KeyError:
            item = None
        return item

    def get_left(self, left_item):
        return self.__left.get(left_item)

    def get_right(self, right_item):
        return self.__right.get(right_item)

    def pop_left(self, left_item, default=None):
        right_item = self.__left.pop(left_item, default)
        self.__right.pop(right_item, default)
        return right_item

    def pop_right(self, right_item, default=None):
        left_item = self.__left.pop(right_item, default)
        self.__left.pop(left_item, default)
        return left_item

    def left_keys(self):
        return self.__left.keys()

    def right_keys(self):
        return self.__right.keys()

    def left_values(self):
        return self.__left.values()

    def right_values(self):
        return self.__right.values()

    def left_items(self):
        return self.__left.items()

    def right_items(self):
        return self.__right.items()

    def clear(self):
        self.__left.clear()
        self.__right.clear()


class WeakList(list):
    """
    List containing weakly referenced items.

    All objects stored in a weak list are only weakly referenced, but
    accessing an element returns a strong reference. If an object referenced
    by an element in a weak list dies, the corresponding weak reference is
    removed automatically.

    :note: only weakly referenceable objects can be stored
    """

    def __init__(self, sequence=None):
        list.__init__(self)
        if not sequence is None:
            self.extend(sequence)
        self.__cmp_key = functools.cmp_to_key(self.__cmp_items)

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            wr = [self.__get_weakref(val) for val in value]
        else:
            wr = self.__get_weakref(value)
        list.__setitem__(self, index, wr)

    def __getitem__(self, index):
        if isinstance(index, slice):
            val = [list.__getitem__(self, item)()
                   for item in range(index.start, index.stop)]
        else:
            val = list.__getitem__(self, index)()
        return val

    def __setslice__(self, start, stop, sequence):
        list.__setslice__(self, start, stop,
                          [self.__get_weakref(el) for el in sequence])

    def __getslice__(self, start, stop):
        return [item() for item in list.__getslice__(self, start, stop) ]

    def __contains__(self, value):
        return list.__contains__(self, self.__get_weakref(value))

    def __add__(self, sequence):
        new_weak_list = WeakList(sequence)
        new_weak_list.extend(self)
        return new_weak_list

    def __iter__(self):
        if len(self) == 0:
            raise StopIteration
        else:
            cnt = 0
            while cnt < len(self):
                yield self.__getitem__(cnt)
                cnt += 1

    def append(self, value):
        list.append(self, self.__get_weakref(value))

    def count(self, value):
        return list.count(self, self.__get_weakref(value))

    def index(self, value):
        return list.index(self, self.__get_weakref(value))

    def insert(self, position, value):
        list.insert(self, position, self.__get_weakref(value))

    def pop(self):
        item_weakref = list.pop(self)
        return item_weakref()

    def remove(self, value):
        self.__delitem__(self.index(value))

    def extend(self, sequence):
        list.extend(self, self.__sequence_to_weakrefs(sequence))

    def sort(self):
        list.sort(self, key=self.__cmp_key)

    def __get_weakref(self, value):
        return ref(value, self.__remove_by_weakref)

    def __remove_by_weakref(self, weakref):
        # Cleanup callback called on garbage collection.
        while True:
            try:
                self.__delitem__(list.index(self, weakref))
            except ValueError:
                break

    def __sequence_to_weakrefs(self, sequence):
        return [self.__get_weakref(el) for el in sequence]

    def __cmp_items(self, left_ref, right_ref):
        return (left_ref() > right_ref()) - (left_ref() < right_ref())


class WeakOrderedSet(MutableSet):
    """
    Ordered set storing weak references to its items.

    Based on a recipe by Raymond Hettinger
    (http://code.activestate.com/recipes/576694-orderedset/).
    """
    def __init__(self, iterable=None):
        MutableSet.__init__(self)
        self.end = end = []
        end += [None, end, end] # sentinel node for doubly linked list
        self.map = WeakKeyDictionary() # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, nxt = self.map.pop(key)
            prev[2] = nxt
            nxt[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __eq__(self, other):
        if isinstance(other, WeakOrderedSet):
            is_eq = len(self) == len(other) and list(self) == list(other)
        else:
            is_eq = set(self) == set(other)
        return is_eq


def generative(func):
    """
    Marks an instance method as generative.
    """
    def wrap(inst, *args, **kw):
        clone = type(inst).__new__(type(inst))
        clone.__dict__ = inst.__dict__.copy()
        return func(clone, *args, **kw)
    return update_wrapper(wrap, func)


def truncate(message, limit=500):
    """
    Truncates the message to the given limit length. The beginning and the
    end of the message are left untouched.
    """
    if len(message) > limit:
        trc_msg = ''.join([message[:limit // 2 - 2],
                           ' .. ',
                           message[len(message) - limit // 2 + 2:]])
    else:
        trc_msg = message
    return trc_msg


class TruncatingFormatter(Formatter):
    """
    Formatter that chops excessive logging argument strings to a defined
    limit. Useful e.g. to restrict logging output from request bodies.

    To use, pass a key "output_limit" to the "extra" argument in your
    logging calls. The logging message will then be truncated to the specified
    length.
    """
    def format(self, record):
        if record.args and hasattr(record, 'output_limit'):
            # Truncate all args to the set limit.
            record.args = tuple([truncate(arg, limit=record.output_limit)
                                 for arg in record.args])
        return Formatter.format(self, record)



"""
Get member view.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 11, 2010.
"""
from everest.views.base import GetResourceView

__docformat__ = "reStructuredText en"
__all__ = ['GetMemberView',
           ]


class GetMemberView(GetResourceView):
    """
    View for GET requests on member resources.
    """
    def _prepare_resource(self):
        return self.context



"""
Representer related utilities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 18, 2011.
"""
from everest.representers.interfaces import ICollectionDataElement
from everest.representers.interfaces import ILinkedDataElement
from everest.representers.interfaces import IRepresenterRegistry
from everest.representers.interfaces import IResourceDataElement
from pyramid.compat import NativeIO
from pyramid.compat import iteritems_
from pyramid.threadlocal import get_current_registry
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
import os

__docformat__ = 'reStructuredText en'
__all__ = ['NewRepresenterConfigurationContext',
           'UpdatedRepresenterConfigurationContext',
           'as_representer',
           'data_element_tree_to_string',
           'get_mapping_registry',
           ]


def as_representer(resource, content_type):
    """
    Adapts the given resource and content type to a representer.

    :param resource: resource to adapt.
    :param str content_type: content (MIME) type to obtain a representer for.
    """
    reg = get_current_registry()
    rpr_reg = reg.queryUtility(IRepresenterRegistry)
    return rpr_reg.create(type(resource), content_type)


def get_mapping_registry(content_type):
    """
    Returns the data element registry for the given content type (a Singleton).

    :Note: This only works after a representer for the given content type
        has been created.
    """
    reg = get_current_registry()
    rpr_reg = reg.queryUtility(IRepresenterRegistry)
    return rpr_reg.get_mapping_registry(content_type)


def data_element_tree_to_string(data_element):
    """
    Creates a string representation of the given data element tree.
    """
    # FIXME: rewrite this as a visitor to use the data element tree traverser.
    def __dump(data_el, stream, offset):
        name = data_el.__class__.__name__
        stream.write("%s%s" % (' ' * offset, name))
        offset += 2
        ifcs = provided_by(data_el)
        if ICollectionDataElement in ifcs:
            stream.write("[")
            first_member = True
            for member_data_el in data_el.get_members():
                if first_member:
                    stream.write('%s' % os.linesep + ' ' * offset)
                    first_member = False
                else:
                    stream.write(',%s' % os.linesep + ' ' * offset)
                __dump(member_data_el, stream, offset)
            stream.write("]")
        else:
            stream.write("(")
            if ILinkedDataElement in ifcs:
                stream.write("url=%s, kind=%s, relation=%s" %
                             (data_el.get_url(), data_el.get_kind(),
                              data_el.get_relation()))
            else:
                first_attr = True
                for attr_name, attr_value in iteritems_(data_el.data):
                    if first_attr:
                        first_attr = False
                    else:
                        stream.write(',%s' % os.linesep
                                     + ' ' * (offset + len(name) + 1))
                    if attr_value is None:
                        continue
                    if not IResourceDataElement in provided_by(attr_value):
                        stream.write("%s=%s" % (attr_name, attr_value))
                    else:
                        stream.write("%s=" % attr_name)
                        __dump(attr_value, stream, offset)
            stream.write(')')
    stream = NativeIO()
    __dump(data_element, stream, 0)
    return stream.getvalue()


class _RepresenterConfigurationContext(object):
    """
    Base class for context managers that configure a representer.

    :ivar options: The representer options map to use within the context.
    :ivar attribute_options: The representer attribute options map to use
      within the context.
    """
    def __init__(self, mapped_class, content_type,
                 options=None, attribute_options=None):
        self.options = options
        self.attribute_options = attribute_options
        self.__mapped_class = mapped_class
        self.__content_type = content_type
        self.__mapping = None

    def __enter__(self):
        mp_reg = get_mapping_registry(self.__content_type)
        self.__mapping = mp_reg.find_or_create_mapping(self.__mapped_class)
        cfg = self._make_configuration(self.__mapping.configuration,
                                          self.options,
                                          self.attribute_options)
        self.__mapping.push_configuration(cfg)

    def __exit__(self, ext_type, value, tb):
        self.__mapping.pop_configuration()

    def _make_configuration(self, current_configuration, options,
                            attribute_options):
        raise NotImplementedError('Abstract method.')


class NewRepresenterConfigurationContext(_RepresenterConfigurationContext):
    """
    A context manager that configures a representer with a newly created
    configuration.
    """
    def _make_configuration(self, current_configuration,
                               options, attribute_options):
        cfg_cls = type(current_configuration)
        cfg = cfg_cls(options=options, attribute_options=attribute_options)
        return cfg


class UpdatedRepresenterConfigurationContext(
                                        _RepresenterConfigurationContext):
    """
    A context manager that configures a representer with a copied and updated
    configuration.
    """
    def _make_configuration(self, current_configuration,
                               options, attribute_options):
        new_cfg = current_configuration.copy()
        upd_cfg = type(new_cfg)(options=options,
                                attribute_options=attribute_options)
        new_cfg.update(upd_cfg)
        return new_cfg



"""
PUT member view.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 14, 2011.
"""
from everest.views.base import PutOrPatchResourceView


__docformat__ = 'reStructuredText en'
__all__ = ['PutMemberView',
           ]


class PutMemberView(PutOrPatchResourceView):
    """
    View for PUT requests on member resources.

    The client sends a PUT request to replace a member resource with a
    new, complete representation (cf. PATCH for partial updates). If the
    request is successful, the server responds with status HTTP OK.
    """
    pass


"""
Mapping and mapping registry.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 4, 2012.
"""
from collections import OrderedDict

from pyramid.compat import iteritems_
from pyramid.compat import itervalues_

from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.representers.attributes import MappedAttribute
from everest.representers.attributes import MappedAttributeKey
from everest.representers.config import RepresenterConfiguration
from everest.representers.dataelements import SimpleCollectionDataElement
from everest.representers.dataelements import SimpleLinkedDataElement
from everest.representers.dataelements import SimpleMemberDataElement
from everest.representers.interfaces import IDataElement
from everest.representers.interfaces import IMemberDataElement
from everest.representers.traversal import DataElementBuilderResourceTreeVisitor
from everest.representers.traversal import ResourceTreeTraverser
from everest.resources.attributes import get_resource_class_attributes
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.resources.interfaces import IResourceLink
from everest.resources.link import Link
from everest.resources.staging import create_staging_collection
from everest.resources.utils import get_collection_class
from everest.resources.utils import provides_collection_resource
from everest.resources.utils import provides_member_resource
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['Mapping',
           'MappingRegistry',
           'PruningMapping',
           'SimpleMappingRegistry',
           ]


class Mapping(object):
    """
    Performs configurable resource <-> data element tree <-> representation
    attribute mappings.

    :property mapped_class: The resource class mapped by this mapping.
    :property data_element_class: The data element class for this mapping.
    """
    #: Flag indicating if this mapping should prune the attribute tree
    #: according to the IGNORE settings.
    is_pruning = False

    def __init__(self, mapping_registry, mapped_class, data_element_class,
                 configuration):
        """
        :param configuration: Mapping configuration object.
        """
        self.__mp_reg = mapping_registry
        self.__mapped_cls = mapped_class
        self.__is_collection_mapping = \
                                provides_collection_resource(mapped_class)
        self.__de_cls = data_element_class
        # List of configurations; the last one added is the active one.
        self.__configurations = [configuration]
        #
        self.__mapped_attr_cache = {}

    def clone(self, options=None, attribute_options=None):
        """
        Returns a clone of this mapping that is configured with the given
        option and attribute option dictionaries.

        :param dict options: Maps representer options to their values.
        :param dict attribute_options: Maps attribute names to dictionaries
          mapping attribute options to their values.
        """
        copied_cfg = self.__configurations[-1].copy()
        upd_cfg = type(copied_cfg)(options=options,
                                   attribute_options=attribute_options)
        copied_cfg.update(upd_cfg)
        return self.__class__(self.__mp_reg, self.__mapped_cls,
                              self.__de_cls, copied_cfg)

    def update(self, options=None, attribute_options=None):
        """
        Updates this mapping with the given option and attribute option maps.

        :param dict options: Maps representer options to their values.
        :param dict attribute_options: Maps attribute names to dictionaries
          mapping attribute options to their values.
        """
        attr_map = self.__get_attribute_map(self.__mapped_cls, None, 0)
        for attributes in attribute_options:
            for attr_name in attributes:
                if not attr_name in attr_map:
                    raise AttributeError('Trying to configure non-existing '
                                         'resource attribute "%s"'
                                         % (attr_name))
        cfg = RepresenterConfiguration(options=options,
                                       attribute_options=attribute_options)
        self.configuration.update(cfg)

    @property
    def configuration(self):
        """
        Returns this mapping's current configuration object.
        """
        # We clear the cache every time the configuration is accessed since
        # we can not guarantee that it stays unchanged.
        self.__mapped_attr_cache.clear()
        return self.__configurations[-1]

    def get_attribute_map(self, mapped_class=None, key=None):
        """
        Returns an ordered map of the mapped attributes for the given mapped
        class and attribute key.

        :param key: Tuple of attribute names specifying a path to a nested
          attribute in a resource tree. If this is not given, all attributes
          in this mapping will be returned.
        """
        if mapped_class is None:
            mapped_class = self.__mapped_cls
        if key is None:
            key = MappedAttributeKey(())
        return OrderedDict([(attr.resource_attr, attr)
                            for attr in self._attribute_iterator(mapped_class,
                                                                 key)])

    def get_attribute(self, attribute_name, mapped_class=None, key=None):
        """
        Returns the specified attribute from the map of all mapped attributes
        for the given mapped class and attribute key. See
        :method:`get_attribute_map` for details.
        """
        attr_map = self.__get_attribute_map(mapped_class, key, 0)
        try:
            return attr_map[attribute_name]
        except KeyError:
            raise AttributeError(attribute_name)

    def has_attribute(self, attribute_name):
        """
        Checks if this mapping has an attribute of the given name.

        :returns: Check result (Boolean)
        """
        return attribute_name in self.__get_attribute_map(None, None, 0)

    def get_attribute_by_repr(self, attribute_repr_name, mapped_class=None,
                              key=None):
        """
        Returns the attribute (specified by its representation name) from
        the map of all mapped attributes for the given mapped class and
        attribute key. See :method:`get_attribute_map` for details.
        """
        attr_map = self.__get_attribute_map(mapped_class, key, 1)
        try:
            return attr_map[attribute_repr_name]
        except KeyError:
            raise AttributeError(attribute_repr_name)

    def has_attribute_repr(self, attribute_repr_name):
        """
        Checks if this mapping has an attribute of the given representation
        name.

        :returns: Check result (Boolean)
        """
        return attribute_repr_name in self.__get_attribute_map(None, None, 1)

    def attribute_iterator(self, mapped_class=None, key=None):
        """
        Returns an iterator over all mapped attributes for the given mapped
        class and attribute key. See :method:`get_attribute_map` for details.
        """
        for attr in self._attribute_iterator(mapped_class, key):
            yield attr

    def terminal_attribute_iterator(self, mapped_class=None, key=None):
        """
        Returns an iterator over all terminal mapped attributes for the given
        mapped class and attribute key. See :method:`get_attribute_map` for
        details.
        """
        for attr in self._attribute_iterator(mapped_class, key):
            if attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                yield attr

    def nonterminal_attribute_iterator(self, mapped_class=None, key=None):
        """
        Returns an iterator over all non-terminal mapped attributes for the
        given mapped class and attribute key. See :method:`get_attribute_map`
        for details.
        """
        for attr in self._attribute_iterator(mapped_class, key):
            if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                yield attr

    def create_data_element(self, mapped_class=None):
        """
        Returns a new data element for the given mapped class.

        :returns: object implementing :class:`IResourceDataElement`.
        """
        if not mapped_class is None and mapped_class != self.__mapped_cls:
            mp = self.__mp_reg.find_or_create_mapping(mapped_class)
            data_el = mp.create_data_element()
        else:
            data_el = self.__de_cls.create()
        return data_el

    def create_linked_data_element(self, url, kind, id=None, # pylint: disable=W0622
                                   relation=None, title=None):
        """
        Returns a new linked data element for the given url and kind.

        :param str url: URL to assign to the linked data element.
        :param str kind: kind of the resource that is linked. One of the
          constantes defined by :class:`everest.constants.RESOURCE_KINDS`.
        :returns: object implementing :class:`ILinkedDataElement`.
        """
        mp = self.__mp_reg.find_or_create_mapping(Link)
        return mp.data_element_class.create(url, kind, id=id,
                                            relation=relation, title=title)

    def create_data_element_from_resource(self, resource):
        """
        Returns a new data element for the given resource object.

        :returns: object implementing :class:`IResourceDataElement`.
        """
        mp = self.__mp_reg.find_or_create_mapping(type(resource))
        return mp.data_element_class.create_from_resource(resource)

    def create_linked_data_element_from_resource(self, resource):
        """
        Returns a new linked data element for the given resource object.

        :returns: object implementing :class:`ILinkedDataElement`.
        """
        mp = self.__mp_reg.find_or_create_mapping(Link)
        return mp.data_element_class.create_from_resource(resource)

    def map_to_resource(self, data_element, resource=None):
        """
        Maps the given data element to a new resource or updates the given
        resource.

        :raises ValueError: If :param:`data_element` does not provide
          :class:`everest.representers.interfaces.IDataElement`.
        """
        if not IDataElement.providedBy(data_element): # pylint:disable=E1101
            raise ValueError('Expected data element, got %s.' % data_element)
        if resource is None:
            coll = \
                create_staging_collection(data_element.mapping.mapped_class)
            agg = coll.get_aggregate()
            agg.add(data_element)
            if IMemberDataElement.providedBy(data_element): # pylint: disable=E1101
                ent = next(iter(agg))
                resource = \
                    data_element.mapping.mapped_class.create_from_entity(ent)
            else:
                resource = coll
        else:
            resource.update(data_element)
        return resource

    def map_to_data_element(self, resource):
        """
        Maps the given resource to a data element tree.
        """
        trv = ResourceTreeTraverser(resource, self.as_pruning())
        visitor = DataElementBuilderResourceTreeVisitor(self)
        trv.run(visitor)
        return visitor.data_element

    def as_pruning(self):
        """
        Returns a clone of this mapping with the `is_pruning` flag set to
        *True*.
        """
        return PruningMapping(self.__mp_reg, self.__mapped_cls, self.__de_cls,
                              self.__configurations[-1])

    def push_configuration(self, configuration):
        """
        Pushes the given configuration object on the stack of configurations
        managed by this mapping and makes it the active configuration.
        """
        self.__mapped_attr_cache.clear()
        self.__configurations.append(configuration)

    def pop_configuration(self):
        """
        Pushes the currently active configuration from the stack of
        configurations managed by this mapping.

        :raises IndexError: If there is only one configuration in the stack.
        """
        if len(self.__configurations) == 1:
            raise IndexError('Can not pop the last configuration from the '
                             'stack of configurations.')
        self.__configurations.pop()
        self.__mapped_attr_cache.clear()

    def with_updated_configuration(self, options=None,
                                   attribute_options=None):
        """
        Returns a context in which this mapping is updated with the given
        options and attribute options.
        """
        new_cfg = self.__configurations[-1].copy()
        if not options is None:
            for o_name, o_value in iteritems_(options):
                new_cfg.set_option(o_name, o_value)
        if not attribute_options is None:
            for attr_name, ao_opts in iteritems_(attribute_options):
                for ao_name, ao_value in iteritems_(ao_opts):
                    new_cfg.set_attribute_option(attr_name, ao_name, ao_value)
#        upd_cfg = type(new_cfg)(options=options,
#                                attribute_options=attribute_options)
#        new_cfg.update(upd_cfg)
        return MappingConfigurationContext(self, new_cfg)

    @property
    def mapped_class(self):
        return self.__mapped_cls

    @property
    def data_element_class(self):
        return self.__de_cls

    @property
    def mapping_registry(self):
        return self.__mp_reg

    def _attribute_iterator(self, mapped_class, key):
        """
        Returns an iterator over the attributes in this mapping for the
        given mapped class and attribute key.

        If this is a pruning mapping, attributes that are ignored because
        of a custom configuration or because of the default ignore rules
        are skipped.
        """
        for attr in \
          itervalues_(self.__get_attribute_map(mapped_class, key, 0)):
            if self.is_pruning:
                do_ignore = attr.should_ignore(key)
            else:
                do_ignore = False
            if not do_ignore:
                yield attr

    def __get_attribute_map(self, mapped_class, key, index):
        if mapped_class is None:
            mapped_class = self.__mapped_cls
        if key is None:
            key = MappedAttributeKey(()) # Top level access.
        attr_maps = self.__mapped_attr_cache.get((mapped_class, key))
        if attr_maps is None:
            attr_maps = self.__cache_attributes(mapped_class, key)
        return attr_maps[index]

    def __cache_attributes(self, mapped_class, key):
        attr_map = self.__collect_mapped_attributes(mapped_class, key)
        # For lookup by repr attribute name, we keep another map.
        repr_attr_map = dict([(attr.repr_name, attr)
                              for attr in itervalues_(attr_map)])
        self.__mapped_attr_cache[(mapped_class, key)] = (attr_map,
                                                         repr_attr_map)
        return (attr_map, repr_attr_map)

    def __collect_mapped_attributes(self, mapped_class, key):
        if isinstance(key, MappedAttributeKey):
            names = key.names
        else:
            names = key
        collected_mp_attrs = OrderedDict()
        is_mapped_cls = mapped_class is self.__mapped_cls
        if len(names) == 0 and is_mapped_cls:
            # Bootstrapping: fetch resource attributes and create new
            # mapped attributes.
            rc_attrs = get_resource_class_attributes(self.__mapped_cls)
            for rc_attr in itervalues_(rc_attrs):
                attr_key = names + (rc_attr.resource_attr,)
                attr_mp_opts = \
                    self.__configurations[-1].get_attribute_options(attr_key)
                new_mp_attr = MappedAttribute(rc_attr, options=attr_mp_opts)
                collected_mp_attrs[new_mp_attr.resource_attr] = new_mp_attr
        else:
            # Indirect access - fetch mapped attributes from some other
            # class' mapping and clone.
            if is_mapped_cls:
                mp = self
            elif len(names) == 0 and self.__is_collection_mapping:
                if provides_member_resource(mapped_class):
                    # Mapping a polymorphic member class.
                    mapped_coll_cls = get_collection_class(mapped_class)
                else:
                    # Mapping a derived collection class.
                    mapped_coll_cls = mapped_class
                mp = self.__mp_reg.find_or_create_mapping(mapped_coll_cls)
            else:
                mp = self.__mp_reg.find_or_create_mapping(mapped_class)
            mp_attrs = mp.get_attribute_map()
            for mp_attr in itervalues_(mp_attrs):
                attr_key = names + (mp_attr.name,)
                attr_mp_opts = \
                    dict(((k, v)
                          for (k, v) in
                            iteritems_(self.__configurations[-1]
                                       .get_attribute_options(attr_key))
                          if not v is None))
                clnd_mp_attr = mp_attr.clone(options=attr_mp_opts)
                collected_mp_attrs[mp_attr.resource_attr] = clnd_mp_attr
        return collected_mp_attrs


class MappingConfigurationContext(object):
    def __init__(self, mapping, configuration):
        self.__mapping = mapping
        self.__configuration = configuration

    def __enter__(self):
        self.__mapping.push_configuration(self.__configuration)

    def __exit__(self, ext_type, value, tb):
        self.__mapping.pop_configuration()


class PruningMapping(Mapping):
    """
    Specialized mapping that applies ignore rules when iterating over
    mapped attributes.
    """
    is_pruning = True


class MappingRegistry(object):
    """
    The mapping registry manages resource attribute mappings by resource
    class.
    """
    member_data_element_base_class = None
    collection_data_element_base_class = None
    linked_data_element_base_class = None
    configuration_class = None
    mapping_class = Mapping

    def __init__(self):
        self.__configuration = self.configuration_class() # pylint: disable=E1102
        self.__mappings = {}
        self.__is_initialized = False

    def _initialize(self):
        # Implement this for static initializations.
        raise NotImplementedError('Abstract method.')

    def set_default_config_option(self, name, value):
        """
        Sets the option of the given name to the given value in the
        registry configuration. This setting then serves as the default for
        future mappings created by this registry.
        """
        self.__configuration.set_option(name, value)

    def create_mapping(self, mapped_class, configuration=None):
        """
        Creates a new mapping for the given mapped class and representer
        configuration.

        :param configuration: configuration for the new data element class.
        :type configuration: :class:`RepresenterConfiguration`
        :returns: newly created instance of :class:`Mapping`
        """
        cfg = self.__configuration.copy()
        if not configuration is None:
            cfg.update(configuration)
        provided_ifcs = provided_by(object.__new__(mapped_class))
        if IMemberResource in provided_ifcs:
            base_data_element_class = self.member_data_element_base_class
        elif ICollectionResource in provided_ifcs:
            base_data_element_class = self.collection_data_element_base_class
        elif IResourceLink in provided_ifcs:
            base_data_element_class = self.linked_data_element_base_class
        else:
            raise ValueError('Mapped class for data element class does not '
                             'implement one of the required interfaces.')
        name = "%s%s" % (mapped_class.__name__,
                         base_data_element_class.__name__)
        de_cls = type(name, (base_data_element_class,), {})
        mp = self.mapping_class(self, mapped_class, de_cls, cfg)
        # Set the data element class' mapping.
        # FIXME: This looks like a hack.
        de_cls.mapping = mp
        return mp

    def set_mapping(self, mapping):
        """
        Registers the given mapping, using the mapped class as key.

        :param mapping: mapping
        :type mapping: :class:`Mapping`
        """
        self.__mappings[mapping.mapped_class] = mapping

    def find_mapping(self, mapped_class):
        """
        Returns the mapping registered for the given mapped class or any of
        its base classes. Returns `None` if no mapping can be found.

        :param mapped_class: mapped type
        :type mapped_class: type
        :returns: instance of :class:`Mapping` or `None`
        """
        if not self.__is_initialized:
            self.__is_initialized = True
            self._initialize()
        mapping = None
        for base_cls in mapped_class.__mro__:
            try:
                mapping = self.__mappings[base_cls]
            except KeyError:
                continue
            else:
                break
        return mapping

    def find_or_create_mapping(self, mapped_class):
        """
        First calls :meth:`find_mapping` to check if a mapping for the given
        mapped class or any of its base classes has been created. If not, a
        new one is created with a default configuration, registered
        automatically and returned.
        """
        mapping = self.find_mapping(mapped_class)
        if mapping is None:
            mapping = self.create_mapping(mapped_class)
            self.set_mapping(mapping)
        return mapping

    def get_mappings(self):
        """
        Returns an iterator over all registered mappings.

        :returns: iterator yielding tuples containing a mapped class as the
          first and a :class:`Mapping` instance as the second item.
        """
        return itervalues_(self.__mappings)


class SimpleMappingRegistry(MappingRegistry):
    """
    Default implementation for a mapping registry using default data element
    and configuration classes.
    """
    member_data_element_base_class = SimpleMemberDataElement
    collection_data_element_base_class = SimpleCollectionDataElement
    linked_data_element_base_class = SimpleLinkedDataElement
    configuration_class = RepresenterConfiguration

    def _initialize(self):
        # Create and register the linked data element class.
        configuration = self.configuration_class()
        mapping = self.create_mapping(Link, configuration)
        self.set_mapping(mapping)




"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 13, 2013.
"""
import pytest

from everest.entities.base import Entity
from everest.querying.specifications import asc
from everest.querying.specifications import eq
from everest.repositories.memory.cache import EntityCache
from everest.repositories.memory.cache import EntityCacheMap
from everest.repositories.memory.querying import EvalFilterExpression
from everest.repositories.memory.querying import EvalOrderExpression
from everest.repositories.state import EntityState
from everest.resources.descriptors import terminal_attribute


__docformat__ = 'reStructuredText en'
__all__ = ['TestEntityCache',
           'TestEntityCacheMap',
           ]


class TestEntityCache(object):
    def test_basics(self):
        ent = MyEntity(id=0)
        cache = EntityCache(entities=[])
        cache.add(ent)
        assert cache.get_by_id(ent.id) is ent
        assert cache.has_id(ent.id)
        assert cache.get_by_slug(ent.slug) is ent
        assert cache.has_slug(ent.slug)
        assert len(cache.get_all()) == 1
        # Adding the same entity twice should not have any effect.
        cache.add(ent)
        assert cache.get_by_id(ent.id) is ent
        assert len(cache.get_all()) == 1
        #
        ent1 = MyEntity(id=0)
        txt = 'FROBNIC'
        ent1.text = txt
        cache.update(EntityState.get_state_data(ent1), ent)
        assert cache.get_by_id(ent.id).text == txt
        assert cache.get_all() == [ent]
        assert list(cache.retrieve()) == [ent]
        cache.remove(ent)
        assert cache.get_by_id(ent.id) is None
        assert cache.get_by_slug(ent.slug) is None

    def test_filter_order_slice(self, configurator):
        configurator.begin()
        try:
            ent0 = MyEntity(id=0)
            ent1 = MyEntity(id=1)
            ent2 = MyEntity(id=2)
            cache = EntityCache(entities=[])
            cache.add(ent0)
            cache.add(ent1)
            cache.add(ent2)
            filter_expr = EvalFilterExpression(~eq(id=0))
            order_expr = EvalOrderExpression(asc('id'))
            slice_key = slice(1, 2)
            assert list(cache.retrieve(filter_expression=filter_expr,
                                       order_expression=order_expr,
                                       slice_key=slice_key)) \
                   == [ent2]
        finally:
            configurator.end()

    def test_allow_none_id_false(self):
        ent = MyEntity()
        cache = EntityCache(entities=[], allow_none_id=False)
        with pytest.raises(ValueError):
            cache.add(ent)


class TestEntityCacheMap(object):
    def test_basics(self):
        ecm = EntityCacheMap()
        ent = MyEntity(id=0)
        ecm.add(MyEntity, ent)
        assert ecm.has_key(MyEntity)
        assert ecm[MyEntity].get_by_id(ent.id) == ent
        assert ent in ecm
        assert list(ecm.keys()) == [MyEntity]
        ecm.remove(MyEntity, ent)
        assert not ent in ecm


class MyEntity(Entity):
    __everest_attributes__ = dict(text=terminal_attribute(str, 'text'))
    text = None



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 16, 2013.
"""
import pytest

from everest.attributes import get_attribute_cardinality
from everest.attributes import is_terminal_attribute
from everest.constants import CARDINALITY_CONSTANTS
from everest.resources.base import Member
from everest.resources.descriptors import member_attribute
from everest.resources.descriptors import terminal_attribute


__docformat__ = 'reStructuredText en'
__all__ = ['TestAttributes',
           ]


class TestAttributes(object):
    @pytest.mark.parametrize('attr_name', ['attr'])
    def test_get_attribute_cardinality(self, attr_name):
        mb_attr = member_attribute(Member, attr_name)
        assert get_attribute_cardinality(mb_attr) == CARDINALITY_CONSTANTS.ONE
        t_attr = terminal_attribute(int, attr_name)
        with pytest.raises(ValueError):
            get_attribute_cardinality(t_attr)

    @pytest.mark.parametrize('attr_name', ['attr'])
    def test_is_terminal_attribute(self, attr_name):
        mb_attr = member_attribute(Member, attr_name)
        assert is_terminal_attribute(mb_attr) is False
        t_attr = terminal_attribute(int, attr_name)
        assert is_terminal_attribute(t_attr) is True



"""
Interfaces for entity and aggregate classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 3, 2011.
"""
from zope.interface import Attribute # pylint: disable=E0611,F0401
from zope.interface import Interface # pylint: disable=E0611,F0401

__docformat__ = "reStructuredText en"
__all__ = ['IAggregate',
           'IEntity',
           ]


# begin interfaces pylint: disable=W0232, E0213, E0211

class IEntity(Interface):
    """
    Marker interface for all entities.
    """

    id = Attribute('Provides a unique ID for this entity instance.')

    slug = Attribute('A unique string identifier for this entity within '
                     'its aggregate. The slug will be used to build URLs.')

    def create_from_data(data):
        """
        """


class IAggregate(Interface):
    """
    Marker interface for all aggregates.

    An aggregate is to an entity what a collection resource is to a member
    resource.
    """

    def create(entity_class, relation=None, ** kw):
        """
        """

    def clone():
        """
        """

    def count():
        """
        """

    def get_by_id(id_key):
        """
        """

    def get_by_slug(slug):
        """
        """

    def iterator():
        """
        """

    def add(entity):
        """
        """

    def remove(entity):
        """
        """

    def query(**options):
        """
        """

# pylint: enable=W0232, E0213, E0211



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 14, 2011.
"""
from everest.entities.utils import slug_from_string
from everest.testing import Pep8CompliantTestCase

__docformat__ = 'reStructuredText en'
__all__ = ['SlugTestCase',
           ]


class SlugTestCase(Pep8CompliantTestCase):

    def test_basic(self):
        str_input = 'spaces with_underscore AND CAPS'
        str_as_slug = 'spaces-with-underscore-and-caps'
        slug = slug_from_string(str_input)
        self.assert_equal(slug, str_as_slug)



"""
Aggregate for the RDB backend.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from everest.entities.base import RootAggregate
from everest.querying.base import EXPRESSION_KINDS

__docformat__ = 'reStructuredText en'
__all__ = ['RdbAggregate',
           ]


class RdbAggregate(RootAggregate):
    """
    Root aggregate implementation for the RDB repository.
    """
    _expression_kind = EXPRESSION_KINDS.SQL

    def query(self):
        # Need to perform a flush here so that filter expressions are always
        # generated correctly. Also, we pass the counting query class to the
        # base class method to optimize paged queries if the backend supports
        # it.
        self._session.flush()
        return RootAggregate.query(
                    self,
                    query_class=self._session_factory.counting_query_class)



"""
Interfaces for representers.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 18, 2011
"""
from zope.interface import Attribute # pylint: disable=E0611,F0401
from zope.interface import Interface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['ICollectionDataElement',
           'ICollectionResourceRepresenter',
           'IDataElement',
           'IDeSerializer',
           'ILinkedDataElement',
           'IMappedClass',
           'IMappingRegistry',
           'IMemberDataElement',
           'IMemberResourceRepresenter',
           'IRepresentationConverter',
           'IRepresenter',
           'IRepresenterRegistry',
           'IResourceDataElement',
           'IResourceRepresenter',
           'ISerializer',
           ]

# interfaces do not provide a constructor. pylint: disable=W0232
# interface methods do not have self pylint: disable=E0213
# interface methods may have no argument: pylint: disable=E0211
class ISerializer(Interface):
    """
    Interface for objects performing serialization.
    """

    def to_string(obj):
        """
        Serializes the given object to a string.
        """

    def to_stream(obj, stream):
        """
        Serializes the given object to the given stream.
        """


class IDeSerializer(Interface):
    """
    Interface for objects performing deserialization.
    """

    def from_string(string_representation):
        """
        Deserializes the given string to an object.
        """

    def from_stream(stream):
        """
        Reads and deserializes the given stream to an object.
        """


class IRepresenter(IDeSerializer, ISerializer):
    """
    Interface for objects serving as representers.

    The representer interface combines the :class:`ISerializer` and
    :class:`IDeSerializer` interfaces. For a given combination of object
    type and MIME content type, a representer knows how to serialize an
    instance of the specified type into a representation of the specified
    content type and how to de-serialize such a representation into an
    object instance.
    """

    # MIME content type of the representation.
    content_type = Attribute('The MIME content type for the handled '
                             'representation.')


class IResourceRepresenter(IRepresenter):
    """
    Marker interface for resource representers.
    """


class IMemberResourceRepresenter(IResourceRepresenter):
    """
    Marker interface for member resource representers.
    """


class ICollectionResourceRepresenter(IResourceRepresenter):
    """
    Marker interface for collection resource representers.
    """


class IMappedClass(Interface):
    """
    Interface for classes with mapped attributes.
    """
    def create_from_data(data_element):
        """
        Creates an instance from the given data element.
        """


class IDataElement(Interface):
    """
    Base interface for data elements.
    """
    #: Mapping responsible for class -> attribute mapping
    mapping = Attribute("Maps classes to attributes")

    def create():
        """
        Factory class method creating a new data element.
        """


class IResourceDataElement(IDataElement):
    """
    Interface for resource data elements.
    """
    def create_from_resource(resource):
        """
        Factory class method creating a new data element from the given
        resource.
        """


class IMemberDataElement(IResourceDataElement):
    """
    Interface for member data elements.
    """
    converter_registry = Attribute("Registry of representation<->value "
                                   "converters.")

    def iterator():
        """
        Returns an iterator yielding (attribute repr name, attribute value)
        pairs for every attribute set on this data element.
        """

    data = Attribute("Ordered dictionary constructed from the return values "
                     "of the :method:`iterator` method.")

    def get_attribute(name):
        """
        Returns the value that was set for the given attribute representation
        name.

        Raises an :class:`AttributeError` if the attribute was not set.
        """

    def set_attribute(name, value):
        """
        Sets the value for the given attribute representation name to the
        given value.

        Raises an :class:`AttributeError` if the underlying mapping does not
        have an attribute with the given representation name.
        """

    def get_terminal(attr):
        """
        Returns the value for the given terminal mapped attribute or `None`
        if the data element does not have a value set for this attribute.
        """

    def set_terminal(attr, value):
        """
        Sets the given mapped terminal attribute to the given value.
        """

    def get_nested(attr):
        """
        Returns the nested data element specified by the given member or
        collection mapped attribute or `None` if the data element does not
        have a value set for this attribute.
        """

    def set_nested(attr, data_element):
        """
        Sets the given mmeber or collection mapped attribute to the given
        data element.
        """


class ICollectionDataElement(IResourceDataElement):
    """
    Interface for collection data elements.
    """
    def add_member(member_data_el):
        """
        Adds the given member data element to this collection data element.
        """

    def get_members():
        """
        Returns all member data elements.
        """

    def __len__():
        """
        Returns the number of member data elements in this collection data
        element.
        """


class ILinkedDataElement(IDataElement):
    """
    Interface for data elements containing a link to a resource.
    """
    def get_url():
        """
        Returns the URL for this data element.
        """

    def get_kind():
        """
        Returns the kind of the resource being linked to (one of the
        constants declared in :class:`everest.constants.RESOURCE_KINDS`).
        """

    def get_id():
        """
        Returns the ID of the resource being linked to.
        """

    def get_relation():
        """
        Returns the relation of the resource being linked to.
        """

    def get_title():
        """
        Returns a title for the resource being linked to.
        """


class IMappingRegistry(Interface):
    def create_mapping(mapped_class, configuration,
                       base_data_element_class=None):
        """
        Creates a new data element class for the given mapped class and
        representer configuration.
        """

    def set_data_element_class(data_element_class):
        """
        Registers the given data element class.
        """

    def get_data_element_class(mapped_class):
        """
        Returns the data element class registered for the given mapped class.
        """

    def get_data_element_classes():
        """
        Returns a list of all registered data element classes.
        """



class IRepresenterRegistry(Interface):
    """
    Marker interface for the representer registry.
    """


class IRepresentationConverter(Interface):
    def from_representation(value):
        """
        Converts the given representation string to a Python value object.
        """

    def to_representation(value):
        """
        Converts the given Python value object into a representation string.
        """

# pylint: enable=W0232,E0213,E0211



"""
Support for staging transient entities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 27, 2013.
"""
from everest.constants import RELATION_OPERATIONS
from everest.entities.base import Aggregate
from everest.entities.base import RelationshipAggregate
from everest.entities.traversal import AruVisitor
from everest.entities.utils import get_entity_class
from everest.querying.base import EXPRESSION_KINDS
from everest.repositories.memory.cache import EntityCacheMap
from everest.resources.utils import get_collection_class
from everest.traversal import SourceTargetDataTreeTraverser

__docformat__ = 'reStructuredText en'
__all__ = ['StagingAggregate',
           'create_staging_collection',
           ]


class StagingAggregate(Aggregate):
    """
    Staging aggregate.

    A staging aggregate is used to build up a new set of entities of the same
    type e.g. from a representation. It does not have a session and therefore
    has no way to persist the changes that are made to the entities it holds.
    """

    def __init__(self, entity_class, cache=None):
        Aggregate.__init__(self)
        self.entity_class = entity_class
        if cache is None:
            cache = EntityCacheMap()
        self.__cache_map = cache
        self.__visitor = AruVisitor(entity_class, self.__add,
                                    self.__remove, self.__update)

    def get_by_id(self, id_key):
        return self.__cache_map.get_by_id(self.entity_class, id_key)

    def get_by_slug(self, slug):
        return self.__cache_map.get_by_slug(self.entity_class, slug)

    def add(self, entity):
        trv = SourceTargetDataTreeTraverser.make_traverser(
                                                entity,
                                                None,
                                                RELATION_OPERATIONS.ADD,
                                                accessor=self)
        trv.run(self.__visitor)

    def remove(self, entity):
        trv = SourceTargetDataTreeTraverser.make_traverser(
                                                None,
                                                entity,
                                                RELATION_OPERATIONS.REMOVE,
                                                accessor=self)
        trv.run(self.__visitor)

    def update(self, entity, target=None):
        trv = SourceTargetDataTreeTraverser.make_traverser(
                                                entity,
                                                target,
                                                RELATION_OPERATIONS.UPDATE,
                                                accessor=self)
        trv.run(self.__visitor)
        return self.__visitor.root

    def query(self):
        return self.__cache_map.query(self.entity_class)

    def __add(self, entity_class, entity):
        self.__cache_map.add(entity_class, entity)

    def __remove(self, entity_class, entity):
        self.__cache_map.remove(entity_class, entity)

    def __update(self, entity_class, source_data, target_entity):
        self.__cache_map.update(entity_class, source_data, target_entity)

    @property
    def expression_kind(self):
        return EXPRESSION_KINDS.EVAL

    def get_root_aggregate(self, rc):
        ent_cls = get_entity_class(rc)
        return StagingAggregate(ent_cls, cache=self.__cache_map)

    def make_relationship_aggregate(self, relationship):
        return RelationshipAggregate(self, relationship)


def create_staging_collection(resource):
    """
    Helper function to create a staging collection for the given registered
    resource.

    :param resource: registered resource
    :type resource: class implementing or instance providing or subclass of
        a registered resource interface.
    """
    ent_cls = get_entity_class(resource)
    coll_cls = get_collection_class(resource)
    agg = StagingAggregate(ent_cls)
    return coll_cls.create_from_aggregate(agg)



"""
Post collection view.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 14, 2011.
"""
from pyramid.httpexceptions import HTTPCreated

from everest.resources.utils import provides_member_resource
from everest.resources.utils import provides_resource
from everest.views.base import ModifyingResourceView


__docformat__ = 'reStructuredText en'
__all__ = ['PostCollectionView',
           ]


class PostCollectionView(ModifyingResourceView):
    """
    View for POST requests on collection resources.

    The client POSTs a representation of the member to the URI of the
    collection. If the new member resource was created successfully, the
    server responds with a status code of HTTP CREATED (201) and a Location
    header that contains the IRI of the newly created resource and a
    representation of it in the body of the response.
    """
    def _process_request_data(self, data):
        if not provides_resource(data):
            rpr = self._get_request_representer()
            resource = rpr.resource_from_data(data)
        else:
            resource = data
        data_is_member = provides_member_resource(resource)
        if data_is_member:
            new_members = [resource]
        else:
            new_members = resource
        was_created = True
        sync_with_repo = False
        for new_member in new_members:
            name_is_none = new_member.__name__ is None
            sync_with_repo |= name_is_none
            if not name_is_none \
               and not self.context.get(new_member.__name__) is None:
                # We have a member with the same name - 409 Conflict.
                result = self._handle_conflict(new_member.__name__)
                was_created = False
                break
            else:
                self.context.add(new_member)
        if was_created:
            if sync_with_repo:
                # This is not pretty, but necessary: When the resource
                # name depends on the entity ID, the pending entity needs
                # to be flushed to the repository before we can access
                # the ID.
                self.context.get_aggregate().sync_with_repository()
            self.request.response.status = self._status(HTTPCreated)
            if data_is_member:
                loc_rc = resource
            else:
                loc_rc = self.context
            self._update_response_location_header(loc_rc)
            result = self._get_result(resource)
        return result



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2012.
"""
from everest.entities.base import Entity
from everest.testing import Pep8CompliantTestCase

__docformat__ = 'reStructuredText en'
__all__ = ['EntitiesTestCase',
           ]


class EntitiesTestCase(Pep8CompliantTestCase):
    def test_base(self):
        ent00 = MyEntity(id=0)
        ent01 = MyEntity(id=0)
        ent1 = MyEntity(id=1)
        self.assert_equal(ent00, ent01)
        self.assert_not_equal(ent00, ent1)


class MyEntity(Entity):
    pass



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 1, 2011.
"""
from everest.resources.base import Member
from everest.resources.descriptors import attribute_alias
from everest.resources.descriptors import collection_attribute
from everest.resources.descriptors import member_attribute
from everest.resources.descriptors import terminal_attribute
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.interfaces import IMyEntityGrandchild
from everest.tests.complete_app.interfaces import IMyEntityParent
import datetime
from everest.constants import CARDINALITIES

__docformat__ = 'reStructuredText en'
__all__ = ['MyEntityChildMember',
           'MyEntityMember',
           'MyEntityGrandchildMember',
           'MyEntityParentMember',
           ]


class MyEntityParentMember(Member):
    relation = 'http://test.org/myentity-parent'
    # String terminal.
    text = terminal_attribute(str, 'text')
    # String terminal with different name in entity.
    text_rc = terminal_attribute(str, 'text_ent')
    #
    text_alias = attribute_alias('text')


class MyEntityMember(Member):
    relation = 'http://test.org/myentity'
    # Member.
    parent = member_attribute(IMyEntityParent, 'parent',
                              cardinality=CARDINALITIES.ONETOONE,
                              backref='child')
    # Collection.
    children = collection_attribute(IMyEntityChild, 'children',
                                    backref='parent')
    # String terminal.
    text = terminal_attribute(str, 'text')
    # String terminal with different name in entity.
    text_rc = terminal_attribute(str, 'text_ent')
    # Number terminal.
    number = terminal_attribute(int, 'number')
    # Date time terminal.
    date_time = terminal_attribute(datetime.datetime, 'date_time')
    # Dotted attribute.
    parent_text = terminal_attribute(str, 'parent.text_ent')


class MyEntityChildMember(Member):
    relation = 'http://test.org/myentity-child'
    # Member.
    parent = member_attribute(IMyEntity, 'parent', backref='children')
    # Collection accessed as entity attribute and represented as
    # "parent equal to parent member" (backreferencing) specification.
    children = collection_attribute(IMyEntityGrandchild,
                                    entity_attr='children',
                                    backref='parent')
    # String terminal.
    text = terminal_attribute(str, 'text')
    # String terminal with different name in entity.
    text_rc = terminal_attribute(str, 'text_ent')


class MyEntityGrandchildMember(Member):
    relation = 'http://test.org/myentity-grandchild'
    # String terminal.
    text = terminal_attribute(str, 'text')
    # String terminal with different name in entity.
    text_rc = terminal_attribute(str, 'text_ent')
    # Member.
    parent = member_attribute(IMyEntityChild, 'parent', backref='children')



"""
Attribute descriptors for resource classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 19, 2011.
"""
from everest.constants import CARDINALITIES
from everest.constants import DEFAULT_CASCADE
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.entities.interfaces import IEntity
from everest.entities.relationship import DomainRelationship
from everest.relationship import RELATIONSHIP_DIRECTIONS
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IResource
from everest.resources.interfaces import IResourceAttribute
from everest.resources.relationship import ResourceRelationship
from everest.resources.utils import get_member_class
from everest.utils import get_nested_attribute
from everest.utils import id_generator
from everest.utils import set_nested_attribute
from zope.interface import implementer # pylint: disable=E0611,F0401
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from zope.interface.interfaces import IInterface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['attribute_alias',
           'attribute_base',
           'collection_attribute',
           'member_attribute',
           'terminal_attribute',
           ]


@implementer(IResourceAttribute)
class attribute_base(object):
    """
    Abstract base class for all attribute descriptors.

    :cvar kind: the resource attribute kind
    :ivar attr_type: the type (or interface) of the controlled entity
      attribute.
    :ivar entity_attr: the entity attribute the descriptor references. May
      be *None*.
    :ivar int index: unique sequential numeric ID for this attribute. Since
      this index is incremented each time a new resource attribute is
      declared, it can be used to establish a well-defined sorting order on
      all attribute declarations of a resource.
    :ivar resource_attr: the resource attribute this descriptor is mapped to.
      This is set after instantiation.
    """
    __index_gen = id_generator()

    #: The resource attribute kind. Set in derived classes.
    kind = None

    def __init__(self, attr_type, entity_attr):
        self.attr_type = attr_type
        self.entity_attr = entity_attr
        self.index = next(self.__index_gen)
        self.resource_attr = None

    def __get__(self, resource, resource_class):
        raise NotImplementedError('Abstract method')

    def __set__(self, resource, value):
        raise NotImplementedError('Abstract method')

    def __str__(self):
        return "%s: rc attr %s, type %s" \
               % (self.__class__.__name__, self.resource_attr, self.attr_type)


class terminal_attribute(attribute_base):
    """
    Descriptor for declaring terminal attributes of a resource as attributes
    from its underlying entity.

    A terminal attribute is an attribute that the framework will not look
    into any further for querying or serialization.
    """
    kind = RESOURCE_ATTRIBUTE_KINDS.TERMINAL

    def __init__(self, attr_type, entity_attr):
        if not isinstance(attr_type, type):
            raise ValueError('The attribute type of a terminal attribute '
                             'must be a class.')
        attribute_base.__init__(self, attr_type, entity_attr)

    def __get__(self, resource, resource_class):
        if resource is None:
            # Class level access.
            obj = self
        else:
            obj = get_nested_attribute(resource.get_entity(), self.entity_attr)
        return obj

    def __set__(self, resource, value):
        set_nested_attribute(resource.get_entity(), self.entity_attr, value)


class _relation_attribute(attribute_base):
    """
    Base class for relation resource descriptors (i.e., descriptors managing
    a related member or collection resource).

    :ivar cardinality: indicates the cardinality of the relationship for
      non-terminal attributes. This is always `None` for terminal attributes.
    :ivar cascade: sets the cascading rules for this relation attribute.
    :ivar resource_backref: attribute of the related resource (relatee) which
      back-references the current resource (relator).
    """
    def __init__(self, attr_type, entity_attr=None, cardinality=None,
                 cascade=DEFAULT_CASCADE, backref=None):
        if not (isinstance(attr_type, type)
                or IInterface in provided_by(attr_type)):
            raise ValueError('The attribute type of a member or collection '
                             ' attribute must be a class or an interface.')
        if entity_attr is None and backref is None:
            raise ValueError('Either the entity_attr or the backref parameter '
                             'to a relation resource attribute may be None, '
                             'but not both.')
        attribute_base.__init__(self, attr_type, entity_attr)
        self.cardinality = cardinality
        self.cascade = cascade
        self.resource_backref = backref
        self.__entity_backref = None

    def __get__(self, resource, resource_class):
        raise NotImplementedError('Abstract method')

    def __set__(self, resource, value):
        raise NotImplementedError('Abstract method')

    def make_relationship(self, relator,
                          direction=
                            RELATIONSHIP_DIRECTIONS.BIDIRECTIONAL):
        """
        Create a relationship object for this attribute from the given
        relator and relationship direction.
        """
        if IEntity.providedBy(relator): # pylint:disable=E1101
            rel = DomainRelationship(relator, self,
                                     direction=direction)
        elif IResource.providedBy(relator): # pylint:disable=E1101
            rel = ResourceRelationship(relator, self,
                                       direction=direction)
        else:
            raise ValueError('Invalid relator argument "%s" for '
                             'relationship; must provide IEntity or '
                             'IResource.' % relator)
        return rel

    @property
    def entity_backref(self):
        if self.__entity_backref is None:
            if self.resource_backref is None:
                self.__entity_backref = None
            else:
                attr_mb_class = get_member_class(self.attr_type)
                backref_rc_descr = getattr(attr_mb_class,
                                           self.resource_backref, None)
#                # We require the backref to be a resource attribute of the
#                # target.
#                if backref_rc_descr is None:
#                    raise ValueError('The "backref" attribute must be a '
#                                     'resource attribute declared on the '
#                                     'target of the descriptor.')
                if not backref_rc_descr is None:
                    self.__entity_backref = backref_rc_descr.entity_attr
                else:
                    self.__entity_backref = self.resource_backref
        return self.__entity_backref


class member_attribute(_relation_attribute):
    """
    Descriptor for declaring member attributes of a resource as attributes
    of its underlying entity.
    """
    kind = RESOURCE_ATTRIBUTE_KINDS.MEMBER

    def __init__(self, attr_type, entity_attr=None,
                 cardinality=CARDINALITIES.MANYTOONE,
                 cascade=DEFAULT_CASCADE, backref=None):
        _relation_attribute.__init__(self, attr_type,
                                     entity_attr=entity_attr,
                                     cardinality=cardinality,
                                     cascade=cascade,
                                     backref=backref)

    def __get__(self, resource, resource_class):
        if not resource is None:
            ent = get_nested_attribute(resource.get_entity(),
                                       self.entity_attr)
            if not ent is None:
                fac = get_member_class(self.attr_type).as_related_member
                member = fac(ent, self.make_relationship(resource))
            else:
                member = None
        else:
            # class level access
            member = self
        return member

    def __set__(self, resource, value):
        if not value is None:
            ent = value.get_entity()
        else:
            ent = None
        set_nested_attribute(resource.get_entity(), self.entity_attr, ent)


class collection_attribute(_relation_attribute):
    """
    Descriptor for declaring collection attributes of a resource as attributes
    from its underlying entity.
    """
    kind = RESOURCE_ATTRIBUTE_KINDS.COLLECTION

    def __init__(self, attr_type, entity_attr=None,
                 cardinality=CARDINALITIES.ONETOMANY,
                 cascade=DEFAULT_CASCADE, backref=None):
        _relation_attribute.__init__(self, attr_type,
                                     entity_attr=entity_attr,
                                     cardinality=cardinality,
                                     cascade=cascade,
                                     backref=backref)

    def __get__(self, member, member_class):
        if not member is None:
            # Create a collection. We can not just return the
            # entity attribute here as that would load the whole entity
            # collection; so we construct a related collection instead.
            parent_coll = member.__parent__
            while not getattr(parent_coll, '__parent__', None) is None \
                  and not ICollectionResource in provided_by(parent_coll):
                parent_coll = parent_coll.__parent__ # pragma: no cover FIXME: test case!
            root_coll = parent_coll.get_root_collection(self.attr_type)
            coll = root_coll.as_related_collection(
                                        root_coll.get_aggregate(),
                                        self.make_relationship(member))
        else:
            # Class level access.
            coll = self
        return coll

    # FIXME: Not sure if we want to support replacement of child containers.
    def __set__(self, member, value):
        raise NotImplementedError('Not implemented.')


class attribute_alias(object):
    """
    Descriptor for declaring an alias to another attribute declared by an
    attribute descriptor.
    """
    def __init__(self, alias_attr):
        self.alias_attr = alias_attr

    def __get__(self, resource, resource_class):
        if resource is None:
            # Class level access.
            obj = self
        else:
            descr = getattr(resource_class, self.alias_attr)
            obj = descr.__get__(resource, resource_class)
        return obj

    def __set__(self, resource, value):
        descr = getattr(type(resource), self.alias_attr)
        descr.__set__(resource, value)
